"""Use aiounifi as a CLI."""

import argparse
import asyncio
from asyncio.timeouts import timeout
import logging
from ssl import SSLContext

import aiohttp

import aiounifi
from aiounifi.controller import Controller
from aiounifi.models.configuration import Configuration

LOGGER = logging.getLogger(__name__)


async def unifi_controller(
    host: str,
    username: str,
    password: str,
    port: int,
    site: str,
    session: aiohttp.ClientSession,
    ssl_context: SSLContext | None = None,
) -> Controller | None:
    """Set up UniFi controller and verify credentials."""
    controller = Controller(
        Configuration(
            session,
            host,
            username=username,
            password=password,
            port=port,
            site=site,
            ssl_context=ssl_context if ssl_context is not None else False,
        )
    )

    try:
        async with timeout(10):
            await controller.login()
        return controller

    except aiounifi.LoginRequired:
        LOGGER.warning("Connected to UniFi at %s but couldn't log in", host)

    except aiounifi.Unauthorized:
        LOGGER.warning("Connected to UniFi at %s but not registered", host)

    except (TimeoutError, aiounifi.RequestError):
        LOGGER.exception("Error connecting to the UniFi controller at %s", host)

    except aiounifi.AiounifiException:
        LOGGER.exception("Unknown UniFi communication error occurred")

    return None


async def main(
    host: str,
    username: str,
    password: str,
    port: int,
    site: str,
    ssl_context: SSLContext | None = None,
) -> None:
    """CLI method for library."""
    LOGGER.info("Starting aioUniFi")

    websession = aiohttp.ClientSession(cookie_jar=aiohttp.CookieJar(unsafe=True))

    controller = await unifi_controller(
        host=host,
        username=username,
        password=password,
        port=port,
        site=site,
        session=websession,
        ssl_context=ssl_context,
    )

    if not controller:
        LOGGER.error("Couldn't connect to UniFi controller")
        await websession.close()
        return

    await asyncio.gather(
        *(
            controller.clients.update(),
            controller.clients_all.update(),
            controller.devices.update(),
            controller.dpi_apps.update(),
            controller.dpi_groups.update(),
            controller.port_forwarding.update(),
            controller.sites.update(),
            controller.system_information.update(),
            controller.traffic_rules.update(),
            controller.traffic_routes.update(),
            controller.vouchers.update(),
            controller.wlans.update(),
        ),
        return_exceptions=True,
    )

    ws_task = asyncio.create_task(controller.start_websocket())

    try:
        while True:
            await asyncio.sleep(1)

    except asyncio.CancelledError:
        pass

    finally:
        ws_task.cancel()
        await websession.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("host", type=str)
    parser.add_argument("username", type=str)
    parser.add_argument("password", type=str)
    parser.add_argument("-p", "--port", type=int, default=8443)
    parser.add_argument("-s", "--site", type=str, default="default")
    parser.add_argument("-D", "--debug", action="store_true")
    args = parser.parse_args()

    LOG_LEVEL = logging.INFO
    if args.debug:
        LOG_LEVEL = logging.DEBUG
    logging.basicConfig(format="%(message)s", level=LOG_LEVEL)

    LOGGER.info(
        "%s, %s, %s, %i, %s",
        args.host,
        args.username,
        args.password,
        args.port,
        args.site,
    )

    try:
        asyncio.run(
            main(
                host=args.host,
                username=args.username,
                password=args.password,
                port=args.port,
                site=args.site,
            )
        )
    except KeyboardInterrupt:
        LOGGER.info("Keyboard interrupt")



"""Test voucher configuration API.

pytest --cov-report term-missing --cov=aiounifi.voucher tests/test_vouchers.py
"""

from datetime import datetime, timedelta

import pytest

from aiounifi.models.voucher import (
    VoucherCreateRequest,
    VoucherDeleteRequest,
)

from .fixtures import VOUCHERS


async def test_voucher_create(mock_aioresponse, unifi_controller, unifi_called_with):
    """Test create voucher."""
    mock_aioresponse.post("https://host:8443/api/s/default/cmd/hotspot", payload={})

    await unifi_controller.request(
        VoucherCreateRequest.create(
            number=1,
            quota=0,
            expire_number=3600,
            expire_unit=1,
            usage_quota=1000,
            rate_max_up=5000,
            rate_max_down=2000,
            note="Unit Testing",
        )
    )

    assert unifi_called_with(
        "post",
        "/api/s/default/cmd/hotspot",
        json={
            "cmd": "create-voucher",
            "n": 1,
            "quota": 0,
            "expire_number": 3600,
            "expire_unit": 1,
            "bytes": 1000,
            "up": 5000,
            "down": 2000,
            "note": "Unit Testing",
        },
    )


async def test_voucher_delete(mock_aioresponse, unifi_controller, unifi_called_with):
    """Test create voucher."""
    mock_aioresponse.post("https://host:8443/api/s/default/cmd/hotspot", payload={})

    await unifi_controller.request(
        VoucherDeleteRequest.create("657e370a4543a555901865c7")
    )

    assert unifi_called_with(
        "post",
        "/api/s/default/cmd/hotspot",
        json={
            "cmd": "delete-voucher",
            "_id": "657e370a4543a555901865c7",
        },
    )


@pytest.mark.usefixtures("_mock_endpoints")
async def test_no_vouchers(unifi_controller, unifi_called_with):
    """Test that no ports also work."""
    vouchers = unifi_controller.vouchers
    await vouchers.update()

    assert unifi_called_with("get", "/api/s/default/stat/voucher")
    assert len(vouchers.values()) == 0


@pytest.mark.parametrize("voucher_payload", [VOUCHERS])
@pytest.mark.usefixtures("_mock_endpoints")
async def test_vouchers(mock_aioresponse, unifi_controller, unifi_called_with):
    """Test that different types of ports work."""
    vouchers = unifi_controller.vouchers
    await vouchers.update()
    assert len(vouchers.values()) == 2

    voucher = vouchers["657e370a4543a555901865c7"]
    assert voucher.id == "657e370a4543a555901865c7"
    assert voucher.site_id == "5a32aa4ee4b0412345678910"
    assert voucher.note == "auto-generated"
    assert voucher.code == "74700-75124"
    assert voucher.quota == 0
    assert voucher.duration == timedelta(minutes=5184000)
    assert voucher.qos_overwrite is False
    assert voucher.qos_usage_quota == 0
    assert voucher.qos_rate_max_up == 0
    assert voucher.qos_rate_max_down == 0
    assert voucher.used == 2
    assert voucher.create_time == datetime.fromtimestamp(1638342818)
    assert voucher.start_time == datetime.fromtimestamp(1638342832)
    assert voucher.end_time == datetime.fromtimestamp(1949382832)
    assert voucher.for_hotspot is False
    assert voucher.admin_name == "Admin"
    assert voucher.status == "USED_MULTIPLE"
    assert voucher.status_expires == timedelta(seconds=244679302)

    voucher = vouchers["61facea3873fdb075ce28d71"]
    assert voucher.id == "61facea3873fdb075ce28d71"
    assert voucher.site_id == "5a32aa4ee4b0412345678910"
    assert voucher.note == ""
    assert voucher.code == "44703-44703"
    assert voucher.quota == 1
    assert voucher.duration == timedelta(minutes=480)
    assert voucher.qos_overwrite is True
    assert voucher.qos_usage_quota == 1000
    assert voucher.qos_rate_max_up == 2000
    assert voucher.qos_rate_max_down == 5000
    assert voucher.used == 0
    assert voucher.create_time == datetime.fromtimestamp(1643826851)
    assert voucher.start_time is None
    assert voucher.end_time is None
    assert voucher.for_hotspot is False
    assert voucher.admin_name == "Admin"
    assert voucher.status == "VALID_ONE"
    assert voucher.status_expires is None

    mock_aioresponse.post(
        "https://host:8443/api/s/default/cmd/hotspot",
        payload={},
        repeat=True,
    )

    await vouchers.create(voucher)
    assert unifi_called_with(
        "post",
        "/api/s/default/cmd/hotspot",
        json={
            "cmd": "create-voucher",
            "n": 1,
            "quota": 1,
            "expire_number": 480,
            "expire_unit": 1,
            "bytes": 1000,
            "up": 2000,
            "down": 5000,
        },
    )

    await vouchers.delete(voucher)
    assert unifi_called_with(
        "post",
        "/api/s/default/cmd/hotspot",
        json={
            "cmd": "delete-voucher",
            "_id": "61facea3873fdb075ce28d71",
        },
    )



"""Hotspot vouchers as part of a UniFi network."""

from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import StrEnum
from typing import NotRequired, Self, TypedDict

from .api import ApiItem, ApiRequest


class VoucherStatus(StrEnum):
    """Voucher status."""

    VALID_ONE = "VALID_ONE"
    VALID_MULTI = "VALID_MULTI"
    USED_MULTIPLE = "USED_MULTIPLE"


class TypedVoucher(TypedDict):
    """Voucher type definition."""

    _id: str
    site_id: str
    note: NotRequired[str]
    code: str
    quota: int
    duration: float
    qos_overwrite: NotRequired[bool]
    qos_usage_quota: NotRequired[str]
    qos_rate_max_up: NotRequired[int]
    qos_rate_max_down: NotRequired[int]
    used: int
    create_time: float
    start_time: NotRequired[float]
    end_time: NotRequired[float]
    for_hotspot: NotRequired[bool]
    admin_name: str
    status: VoucherStatus
    status_expires: float


@dataclass
class VoucherListRequest(ApiRequest):
    """Request object for voucher list."""

    @classmethod
    def create(cls) -> Self:
        """Create voucher list request."""
        return cls(
            method="get",
            path="/stat/voucher",
        )


@dataclass
class VoucherCreateRequest(ApiRequest):
    """Request object for voucher create."""

    @classmethod
    def create(
        cls,
        expire_number: int,
        expire_unit: int = 1,
        number: int = 1,
        quota: int = 0,
        usage_quota: int | None = None,
        rate_max_up: int | None = None,
        rate_max_down: int | None = None,
        note: str | None = None,
    ) -> Self:
        """Create voucher create request.

        :param expire_number: expiration of voucher per expire_unit
        :param expire_unit: scale of expire_number, 1 = minute, 60 = hour, 3600 = day
        :param number: number of vouchers
        :param quota: number of using; 0 = unlimited
        :param usage_quota: quantity of bytes allowed in MB
        :param rate_max_up: up speed allowed in kbps
        :param rate_max_down: down speed allowed in kbps
        :param note: description
        """
        data = {
            "cmd": "create-voucher",
            "n": number,
            "quota": quota,
            "expire_number": expire_number,
            "expire_unit": expire_unit,
        }
        if usage_quota:
            data["bytes"] = usage_quota
        if rate_max_up:
            data["up"] = rate_max_up
        if rate_max_down:
            data["down"] = rate_max_down
        if note:
            data["note"] = note

        return cls(
            method="post",
            path="/cmd/hotspot",
            data=data,
        )


@dataclass
class VoucherDeleteRequest(ApiRequest):
    """Request object for voucher delete."""

    @classmethod
    def create(
        cls,
        obj_id: str,
    ) -> Self:
        """Create voucher delete request."""
        data = {
            "cmd": "delete-voucher",
            "_id": obj_id,
        }
        return cls(
            method="post",
            path="/cmd/hotspot",
            data=data,
        )


class Voucher(ApiItem):
    """Represents a voucher."""

    raw: TypedVoucher

    @property
    def id(self) -> str:
        """ID of voucher."""
        return self.raw["_id"]

    @property
    def site_id(self) -> str:
        """Site ID."""
        return self.raw["site_id"]

    @property
    def note(self) -> str:
        """Note given by creator to voucher."""
        return self.raw.get("note") or ""

    @property
    def code(self) -> str:
        """Code of voucher in known format 00000-00000.

        To enter the code on the captive portal, the hyphen must be placed after the fifth digit.
        """
        c = self.raw.get("code", "")
        # API returns the code without a hyphen. But this is necessary. Separate the API string after the fifth digit.
        return f"{c[:5]}-{c[5:]}"

    @property
    def quota(self) -> int:
        """Allowed uses (0 = unlimited) of voucher."""
        return self.raw["quota"]

    @property
    def duration(self) -> timedelta:
        """Expiration of voucher."""
        return timedelta(minutes=self.raw["duration"])

    @property
    def qos_overwrite(self) -> bool:
        """Defaults for QoS overwritten by the use of this voucher."""
        return self.raw.get("qos_overwrite", False)

    @property
    def qos_usage_quota(self) -> int:
        """Quantity of bytes (in MB) allowed when using this voucher."""
        return int(self.raw.get("qos_usage_quota", 0))

    @property
    def qos_rate_max_up(self) -> int:
        """Up speed (in kbps) allowed when using this voucher."""
        return self.raw.get("qos_rate_max_up", 0)

    @property
    def qos_rate_max_down(self) -> int:
        """Down speed (in kbps) allowed when using this voucher."""
        return self.raw.get("qos_rate_max_down", 0)

    @property
    def used(self) -> int:
        """Number of uses of this voucher."""
        return self.raw["used"]

    @property
    def create_time(self) -> datetime:
        """Create datetime of voucher."""
        return datetime.fromtimestamp(self.raw["create_time"])

    @property
    def start_time(self) -> datetime | None:
        """Start datetime of first usage of voucher."""
        if "start_time" in self.raw:
            return datetime.fromtimestamp(self.raw["start_time"])
        return None

    @property
    def end_time(self) -> datetime | None:
        """End datetime of latest usage of voucher."""
        if "end_time" in self.raw:
            return datetime.fromtimestamp(self.raw["end_time"])
        return None

    @property
    def for_hotspot(self) -> bool:
        """For hotspot use."""
        return self.raw.get("for_hotspot", False)

    @property
    def admin_name(self) -> str:
        """Creator name of voucher."""
        return self.raw["admin_name"]

    @property
    def status(self) -> VoucherStatus:
        """Status of voucher."""
        return self.raw["status"]

    @property
    def status_expires(self) -> timedelta | None:
        """Status expires in seconds."""
        if (status_expiry := self.raw["status_expires"]) > 0:
            return timedelta(seconds=status_expiry)
        return None
